/*
 * Дана последовательность натуральных чисел a0, ...
 * Создать многопоточное приложение для поиска суммы квадратов sum(a[i]^2. 
 * В приложении вычисления должны независимо выполнять четыре потока.
*/

#include <iostream>
#include <iomanip>
#include <limits>
#include <ctime>
#include <pthread.h>

struct Package {
    double* array;  // Указатель на начало области обработки
    int threadNum;  // Номер потока
    double sum;     // Формируемая частичная сумма
} ;

const unsigned int arrSize = 10000000;
//const unsigned int arrSize = 500000000;
//const unsigned int arrSize = 900000000;

double *A ;  //последовательность чисел a0...

//const int threadNumber = 1; // Количество потоков
//const int threadNumber = 2; // Количество потоков
const int threadNumber = 4; // Количество потоков
//const int threadNumber = 8; // Количество потоков

//стартовая функция для дочерних потоков
void *func(void *param) {    //вычисление суммы квадратов в потоке
    // Востановление структуры
    Package* p = (Package*)param;
    p->sum = 0.0;
    for(unsigned int i = p->threadNum ; i < arrSize; i+=threadNumber) {
        p->sum += p->array[i] * p->array[i];
    }
    return nullptr;
}

int main() {
    double rez = 0.0 ; //для записи окончательного результата

    // заполнение массива A
    A = new double[arrSize];
    if(A == nullptr) {
        std::cout << "Incorrect size of vector = " << arrSize << "\n";
        return 1;
    }
    for(int i = 0; i < arrSize; ++i) {
        A[i] = double(i);
    }
    
    pthread_t thread[threadNumber];
    Package pack[threadNumber];
    
    clock_t start_time =  clock(); // начальное время
    
    //создание дочерних потоков
    for (int i=0 ; i<threadNumber ; i++) {
        // Формирование структуры для передачи потоку
        pack[i].array = A;
        pack[i].threadNum = i;
        pthread_create(&thread[i], nullptr, func, (void*)&pack[i]) ;
    }
    
    clock_t thread_started_time = clock(); // конечное время
    
    for (int i = 0 ; i < threadNumber; i++) {    // ожидание завершения работы дочерних потоков
        pthread_join(thread[i], nullptr) ;       // и получение результата их вычислений
        rez += pack[i].sum;
    }

    clock_t end_time = clock(); // конечное время
    
    //вывод результата
    std::cout << "Сумма квадратов = " << rez << "\n" ;
    std::cout << "Сумма квадратов = " << 
        std::setprecision(20) << rez << "\n" ;
    std::cout << "Сумма квадратов = " << std::scientific <<
        std::setprecision(std::numeric_limits<double>::digits10 + 1) << rez << "\n" ; 

    std::cout << "Потоки запустились = " << thread_started_time - start_time << "\n";
    std::cout << "Время счета и сборки = " << end_time - start_time << "\n";

    delete[] A;
    return 0;
}
